document.addEventListener('DOMContentLoaded', function () {
    if (typeof mermaid !== 'undefined') {
      mermaid.initialize({
        themeVariables:{
            lineColor: '#888888'
        }
      })
    }
  })
  